const logger = require.main.require('./common/logger.js'),
    electron = require('electron'),
    { app, BrowserWindow, globalShortcut, ipcMain } = require('electron'),
    { fork, spawn } = require('child_process'),
    fs = require('fs'),
    lkt = require.main.require('./common/lkt.js'),
    isRunning = require('./common/isRunning.js');

let electronEjs = require('electron-ejs');
let prompt = require('electron-prompt');
let ejs = new electronEjs({ app: 'Amadeus' }, {});

const packageConfig = require('./package.json').config;
logger.debug('main', 'Package config is ' + JSON.stringify(packageConfig));

var config = require('./common/config.js');
var db = require.main.require('./common/db.js');
var Tail = require('tail').Tail;

var client; /* Sub process for the express server */

/* Timers functions, needs to be clear on app quit */
var loopIntervals = [];
var loopTimeouts = [];

function clearTimers() {
    loopIntervals.forEach(timer => clearInterval(timer));
    loopTimeouts.forEach(timer => clearTimeout(timer));
}

class Lektor {
    constructor() {
        this.closed = true;
        this.attached = false;
        isRunning({ win: 'lektord.exe', linux: 'lektord', mac: 'lektord' }).then(v => {
            if (v && config.content.lektord.attach) {
                logger.info('main', `Lektord is running! Attach to it`);
                this.attached = true;
            } else if (v) {
                logger.error('error', `Lektord is running and attach mode is disable, exit!`);
                app.quit();
            } else {
                logger.debug('main', `Lektord not running, start it!`);
                this.closed = false;
                this.process = spawn('lektord', ['-F']);
                this.process.stderr.on('data', data => {
                    logger.lektord(data);
                });
                this.process.on('close', code => {
                    logger.warn('main', `Lektor exited with code ${code}`);
                    this.lektor_closed = true;
                    app.quit();
                });
            }
        });
    }

    exit() {
        if (this.closed) {
            logger.warning('main', 'Lektord is already closed');
            return;
        }
        logger.info('main', 'Closing lektord');
        this.closed = true;
        this.process.kill('SIGTERM');
    }
}

/* The last thing we got from the search bar */
var __lastFilter = '';

var win = null; /* The main window */

/**********************
 * Reinit the logfile *
 **********************/

fs.truncate(logger.logfile, 0, () => {});
var tail = new Tail(logger.logfile);
tail.on('line', function (data) {
    console.log(data);
});

/***************************************************************
 * Creates the main window and process for the admin interface *
 ***************************************************************/

function defaultWindowOptions(title) {
    return {
        title: title,
        width: 720,
        height: 360,
        hasShadow: false,
        frame: true,
        menuBarVisible: false,
        webPreferences: {
            nodeIntegration: true,
            worldSafeExecuteJavaScript: true,
            contextIsolation: true,
        },
    };
}

function createInstanceWindow() {
    /* Main window */
    const data = fs.readFileSync('./config.json');
    const json = data.toString('utf8');
    settings = JSON.parse(json);
    const opt = defaultWindowOptions('Amadeus');
    opt.width = 1280;
    opt.height = 720;
    opt.minWidth = 1000;
    opt.minHeight = 360;
    opt.menuBarVisible = false;
    if (!settings.isWindows) {
        opt.frame = false;
    } else {
        opt.frame = true;
        opt.autoHideMenuBar = true;
    }
    opt.webPreferences.contextIsolation = false; /* XXX: Otherwise 'require' is not defined in instance/index.html */
    win = new BrowserWindow(opt);
    win.loadURL(`file://${__dirname}/instance/index.ejs`);
    win.once('ready-to-show', () => win.show());
}

function createDevToolsWindow() {
    var focused = BrowserWindow.getFocusedWindow();
    if (focused) {
        if (!focused.isDevToolsOpened()) {
            focused.webContents.openDevTools({ mode: 'detach' });
        } else {
            focused.webContents.closeDevTools();
        }
    }
}

function createLogWindow() {
    /* Log window */
    logger.debug('main', 'Open log view');
    const opt = defaultWindowOptions('Amadeus - Log window');
    opt.parent = win;
    opt.frame = false;
    opt.webPreferences.contextIsolation = false; /* XXX: Otherwise 'require' is not defined in instance/index.html */
    const logwindow = new BrowserWindow(opt);
    logwindow.loadURL(`file://${__dirname}/instance/log.ejs`);
    logwindow.once('ready-to-show', () => logwindow.show());
}

function createUserViewWindow() {
    /* User view window */
    logger.debug('main', 'Open client view');
    const opt = defaultWindowOptions('Amadeus - User view');
    const uv = new BrowserWindow(opt);
    uv.loadURL(`http://localhost:${config.content.client.port}/`);
    uv.once('ready-to-show', () => uv.show());
}

function createKurisuWindow() {
    /* Kurisu window */
    logger.debug('main', 'Open kurisu view');
    const opt = defaultWindowOptions('Amadeus - Kurisu');
    const kurisu = new BrowserWindow(opt);
    kurisu.loadURL(config.content.kurisu.url);
    kurisu.once('ready-to-show', () => kurisu.show());
}

function setupKaraDB(){
    db.setup_search_engine()
}

/******************************
 * The end of the application *
 ******************************/

ipcMain.on('close-app', () => {
    app.quit();
    clearTimers();
});

app.on('quit', () => {
    logger.info('main', 'Send SIGTERM to express process, lektord and tailler process');
    client.kill('SIGTERM');
    lektor.exit();
});

/*********************************
 * The beggin of the application *
 *********************************/

app.on('ready', () => {
    logger.info('main', 'Main window is ready');
    client = fork('client/main.js');
    lektor = new Lektor();
    createInstanceWindow();

    win.webContents.on('did-finish-load', () => {
        logger.debug('main', 'webContents got "did-finish-load"');
        win.webContents.send('send-lektord-is-attached', lektor.attached);
        win.webContents.on('console-message', (event, level, message, line, sourceId) => {
            switch (level) {
                case 0:
                    logger.debug('console', `[${sourceId}+${line}] ${message.replace(/\n|\r|  /g, '')}`);
                    break;
                case 1:
                    logger.warn('console', `[${sourceId}+${line}] ${message.replace(/\n|\r|  /g, '')}`);
                    break;
                case 2:
                default:
                    logger.error('console', `[${sourceId}+${line}] ${message.replace(/\n|\r|  /g, '')}`);
                    break;
            }
        });
    });

    globalShortcut.register('CommandOrControl+F12', () => createDevToolsWindow());
});

/***************
 * Kurisu View *
 ***************/

ipcMain.on('toggle-kurisu', () => createKurisuWindow());
ipcMain.on('toggle-client-view', () => createUserViewWindow());
ipcMain.on('toggle-log-view', () => createLogWindow());

/*********************************
 * Messages from the main window *
 *********************************/

var songTimeData = { elapsed: 0, total: 100, state: 'stop', song: '0' };
var counterTime = 0;

ipcMain.on('cmd-play', () => lkt.commandPlay().then(arg => logger.debug('main', 'Returned from cmd-play')));
ipcMain.on('cmd-pause', () => lkt.commandPause().then(arg => logger.debug('main', 'Returned from cmd-pause')));
ipcMain.on('cmd-unpause', () => lkt.commandUnpause().then(arg => logger.debug('main', 'Returned from cmd-unpause')));
ipcMain.on('cmd-shuffle', () => lkt.commandShuffle().then(arg => logger.debug('main', 'Returned from cmd-shuffle')));
ipcMain.on('cmd-stop', () => lkt.commandStop().then(arg => logger.debug('main', 'Returned from cmd-stop')));
ipcMain.on('cmd-clear', () => lkt.commandClear().then(arg => logger.debug('main', 'Cleared queue with cmd-clear')));
ipcMain.on('update-database', () => lkt.commandUpdateDatabase());
ipcMain.on('dry-update-database', () => lkt.commandDryUpdateDatabase());

/* Fill the pannel with the content of the DB.
 * The `arg` is the HTML object of the pannel */
ipcMain.on('reload-db-request', (event, arg) => {
    const callback = karas => event.reply('reload-db-responce', karas);
    if (arg && arg.search && arg.search !== '') {
        __lastFilter = arg.search;
        logger.debug('main', `Reload DB with search '${arg}'`);
        karas = db.search(__lastFilter);
        callback(karas);
    } else if ((arg && arg.search === '') || __lastFilter === '') {
        __lastFilter = '';
        callback([]);
    } else {
        logger.debug('main', `Reload DB with last filter '${__lastFilter}'`);
        karas = db.search(__lastFilter);
        callback(karas);
    }
});

ipcMain.on('queue-moved-kara', (event, movement) => {
    if (movement.to != movement.from) {
        lkt.commandMove(movement.from, movement.to);
    }
});

var currentPlaylist = '';

ipcMain.on('play-kara-queue-pos', (event, arg) => lkt.commandPlayPos(arg.position));
ipcMain.on('add-kara-queue-id', (event, arg) => lkt.commandQueueAddId(arg.id));
ipcMain.on('insert-kara-queue-id', (event, arg) => lkt.commandQueueInsertId(arg.id));
ipcMain.on('download-kara-id', (event, arg) => {
    logger.info('download', 'download');
    lkt.commandDownloadId(arg.id);
});
ipcMain.on('delete-kara-queue-pos', (event, arg) => lkt.commandQueueDelPos(arg.position));
ipcMain.on('add-kara-playlist-id', (event, arg) => {
    if (currentPlaylist != '') {
        lkt.commandPlaylistAddId(currentPlaylist, arg.id);
    }
});
ipcMain.on('del-kara-playlist-id', (event, arg) => lkt.commandPlaylistRemoveId(currentPlaylist, arg.id));

ipcMain.on('clear-playlist', (event, arg) => {
    if (currentPlaylist != '') {
        lkt.commandPlaylistClear(currentPlaylist);
    }
});

ipcMain.on('delete-playlist', (event, arg) => {
    if (currentPlaylist != '') {
        lkt.commandPlaylistDelete(currentPlaylist);
    }
});

ipcMain.on('add-current-playlist-queue', (event, arg) => {
    if (currentPlaylist != '') {
        lkt.commandAddPlaylistToQueue(currentPlaylist);
    }
});

ipcMain.on('add-kara-queue-pos', (event, addparams) => {
    lkt.commandQueueAddId(addparams.id).then(() => lkt.commandMove(addparams.queueSize + 1, addparams.position));
});

ipcMain.on('get-playlist-data', (event, playlist) => {
    if (playlist) {
        currentPlaylist = playlist;
        lkt.commandPlaylistListKaras(currentPlaylist).then(karas => {
            win.webContents.send('playlist-data-responce', karas);
        });
    }
});

ipcMain.on('add-kara-queue-search-result', (event, search) => {
    lkt.commandQueueAddSearch(search);
});

ipcMain.on('new-playlist', (event, arg) => {
    prompt({
        title: 'New playlist',
        label: 'Enter playlist name',
        type: 'input',
        customStylesheet: 'style/bootstrap-4.5.2-dist/css/bootstrap.min.css',
    }).then(playlist => {
        if (playlist != null) {
            logger.debug('test', playlist);
            lkt.commandPlaylistCreate(playlist);
        }
    });
});

ipcMain.on('verify-lektord', (event, arg) => {
    lkt.ping().then(sta => {
        logger.debug('main', `Status from ping is ${sta}`);

        /* Install the clear function before registering all timers */
        win.on('close', () => {
            logger.info('main', 'Close instance window');
            clearTimers();
        });

        var destroyed = false;
        win.webContents.on('destroyed', () => {
            logger.debug('main', 'WebContent has been destroyed');
            destroyed = true;
            clearTimers();
        });

        lkt.reloadState();
        setupKaraDB();

        loopTimeouts.concat([
            setTimeout(() => lkt.idleActualisation(), 1000),
            setTimeout(() => lkt.statusActualisation(), 1500),
            setTimeout(() => {
                if (destroyed) return;
                db.queueAll().then(karas => win.webContents.send('reload-queue-responce', karas));
            }, 1000),
        ]);

        loopIntervals.concat([
            setInterval(() => {
                if (!destroyed && lkt.isQueueUpdated()) {
                    lkt.setQueueUpdated(false);
                    db.queueAll().then(karas => win.webContents.send('reload-queue-responce', karas));
                }
            }, 50),
            /* Position in the kara changed? */
            setInterval(() => {
                if (destroyed) return;
                var newSongTimeData = lkt.getSongTimeData();
                if (newSongTimeData.elapsed != songTimeData.elapsed || newSongTimeData.song != songTimeData.song) {
                    songTimeData = newSongTimeData;
                    win.webContents.send('send-song-time-data', {
                        elapsed: songTimeData.elapsed,
                        total: songTimeData.total,
                        song: songTimeData.song,
                        state: songTimeData.state,
                    });
                    counterTime = 0;
                } else if (newSongTimeData.state == 'play') {
                    counterTime++;
                    win.webContents.send('send-song-time-data', {
                        elapsed: songTimeData.elapsed + counterTime / 20.0,
                        total: songTimeData.total,
                        song: songTimeData.song,
                        state: songTimeData.state,
                    });
                } else {
                    win.webContents.send('send-song-time-data', {
                        elapsed: songTimeData.elapsed + counterTime / 20.0,
                        total: songTimeData.total,
                        song: songTimeData.song,
                        state: songTimeData.state,
                    });
                }
            }, 50),
            /* Get running processes */
            setInterval(() => {
                if (destroyed) return;
                isRunning({ win: 'klkt.exe', mac: 'klkt', linux: 'klkt' })
                    .then(vklkt =>
                        isRunning({ win: 'lektord.exe', mac: 'lektord', linux: 'lektord' })
                            .then(vlektord =>
                                win.webContents.send('send-runnings', {
                                    klkt: vklkt,
                                    lektord: vlektord,
                                })
                            )
                            .catch(() => logger.error('main', "Can't know if lektord is running"))
                    )
                    .catch(() => logger.error('main', "Can't know if klkt is running"));
            }, 1000),
            /* Has the status of lektord changed? */
            setInterval(() => {
                if (!destroyed && lkt.isStatusUpdated()) {
                    win.webContents.send('send-state', lkt.getStatus().state);
                    const current = lkt.getCurrent();
                    if (typeof current != 'undefined') {
                        win.webContents.send('send-song-mdt', current);
                    }
                }
            }, 50),
        ]);
        setInterval(() => {
            if (lkt.isPlaylistsUpdated()) {
                lkt.setPlaylistsUpdated(false);
                lkt.commandPlaylistsList().then(playlists => {
                    win.webContents.send('playlists-updated', playlists);
                });
                if (currentPlaylist != '') {
                    lkt.commandPlaylistListKaras(currentPlaylist).then(karas => {
                        win.webContents.send('playlist-data-responce', karas);
                    });
                }
            }
            if (lkt.isDBUpdated()) {
                lkt.setDBUpdated(false);
                win.webContents.send('check-db-updated');
            }
        }, 50);

        setInterval(() => {
            if (currentPlaylist != '' && lkt.isPlaylistsUpdated()) {
                lkt.setPlaylistsUpdated(false);
                lkt.commandPlaylistListKaras(currentPlaylist).then(karas => {
                    win.webContents.send('playlist-data-responce', karas);
                });
            }
        }, 50);
    });
});

ipcMain.on('set-settings', (event, arg) => {
    logger.info('main', `Settings are now: ${JSON.stringify(arg)}`);

    config.content.lektord.host = arg.host;
    config.content.lektord.port = arg.port;
    config.content.lektord.attach = arg.attachmode;

    config.content.database.path = arg.db;

    config.content.kurisu.url = arg.kurisuurl;

    config.content.client.port = arg.clientport;

    /* prettier-ignore */
    {
        config.content.log.lektord = arg.loglektord;
        config.content.log.level   = arg.loglevel.ERROR ?   'error'
                                   : arg.loglevel.WARNING ? 'warning'
                                   : arg.loglevel.INFO ?    'info'
                                   :                        'debug';
    }

    config.flush();
});
